// -*- C++ -*-
//
// eventDisplay.cc
//
// Author:  Mitsuhiro YAMAGA
// Created: Wed Aug 28 10:09:54 JST 2002
//
// $Id: eventDisplay.cc,v 1.6 2003/01/21 08:18:02 yamaga Exp $
//
// Revision History
//
// $Log: eventDisplay.cc,v $
// Revision 1.6  2003/01/21 08:18:02  yamaga
// Added -cycle option.
//
// Revision 1.5  2002/10/15 17:00:07  yamaga
// Added '-tracking' option.
//
// Revision 1.4  2002/10/14 06:20:05  yamaga
// Added -pause option for offline mode.
//
// Revision 1.3  2002/10/07 13:52:18  yamaga
// Added -noscreen option. User_term() called correctly.
//
// Revision 1.2  2002/09/28 05:24:18  yamaga
// user_process_event() returns void.
//
// Revision 1.1  2002/09/27 12:37:02  yamaga
// new package
//
//
//

#include <iostream>
#include <list>
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include "midas.h"
#include "E391Detector/E391Detector.h"
#include "MidasManager/MidasManager.h"


// globals
int     Debug = 0;
int     WkType = 1;
double  Nsigma = 2;
bool    Tracking = false;
bool    Pause = false;
double  Cycle = 2.5;
double  Freq = 1;
int     Interval = 0;
int     Usleep = 0;
double  RotAngle = 0;
double  AdcCutL = 500;
double  AdcCutH = 580;
unsigned int  TriggerBit = 0;
std::string  Psfile;
E391Detector&  e391det = E391Detector::instance();

/* ODB structures */
RUNINFO  runinfo;


//////// user code is defined in the user_ana.cc
extern void  user_init();
extern void  user_begin_of_run( int runnum );
extern void  user_process_event( int runnum,int spillnum,
                                 int eventnum,int trigbit );
extern void  user_process_event_exp( HNDLE hbuf,HNDLE request_id,
                                     EVENT_HEADER* pheader, void* pevent );
extern void  user_end_of_run( int runnum );
extern void  user_term();


////////
// usage
static void
usage( int argc, char* argv[] )
{
  std::cout << "usage : " << argv[0] << " [options]"
            << std::endl 
            << "  options:" << std::endl
            << "    -h <hostname> (connect to the running experiment.)"
            << std::endl 
            << "    -i <filenames...> (you can use %05d for run_number."
            << " e.g. run%05d.mid)"
            << std::endl 
            << "    -r <run#from> <run#to> (work with %05d)"
            << std::endl
            << "    -n <#event>"
            << std::endl 
            << "    -skip <skipevent>"
            << std::endl
            << "    -adccut <low> <high> (ADC cut-off: low-range high-range)"
            << std::endl
            << "    -nsigma <n> (ADC cut-off < mean + n*sigma)"
            << std::endl
            << "    -int <n> (display once per n_event)"
            << std::endl
            << "    -pause (pause until hitting 'enter' key)"
            << std::endl
            << "    -cycle <n> (n second cycle of beam extraction, FOR ONLINE ONLY)"
            << std::endl
            << "    -freq <n> (display n event per sec, FOR ONLINE ONLY)"
            << std::endl
            << "    -usleep <n> (usleep n usec)"
            << std::endl
            << "    -trig <n> (trigger bit)"
            << std::endl
            << "    -rot <n> (rotation angle)"
            << std::endl
            << "    -tracking (tracking the cosmic ray) "
            << std::endl
            << "    -psfile <filename>"
            << std::endl
            << "    -noscreen (Do not display on the screen. Usable with -psfile option.)"
            << std::endl;
}


////////
// main
int
main( int argc, char* argv[] )
{
  if ( argc < 2 ){
    usage( argc,argv );
    return( 1 );
  }

  // command parser
  int  Skipevent = 0;
  int  Nevent  = -1;
  int  runFrom = 0;
  int  runTo   = 0;
  std::list<std::string>  iflist;
  std::list<std::string>  iftmp;
  char  exphost[64];
  for( int i=1;i<argc;i++ ){
    if ( argv[i][0] == '-' ){
      if ( strcmp( argv[i],"-h" ) == 0 ){
        if ( ++i<argc && argv[i][0] != '-' ){
          strcpy( exphost,argv[i] );
        }
        else
          i--;
      }
      else if ( strcmp( argv[i],"-i" ) == 0 ){
        while( ++i<argc && argv[i][0] != '-' ){
          iftmp.push_back( argv[i] );
        }
        i--;
      }
      else if ( strcmp( argv[i],"-r" ) == 0 ){
        runFrom =  strtol( argv[++i],NULL,0 );
        runTo   =  strtol( argv[++i],NULL,0 );
      }
      else if ( strcmp( argv[i],"-n" ) == 0 ){
        Nevent = strtol( argv[++i],NULL,0 );
      }
      else if ( strcmp( argv[i],"-skip" ) == 0 ){
        Skipevent = strtol( argv[++i],NULL,0 );
      }
      else if ( strcmp( argv[i],"-adccut" ) == 0 ){
        AdcCutL = strtod( argv[++i],NULL );
        AdcCutH = strtod( argv[++i],NULL );
      }
      else if ( strcmp( argv[i],"-nsigma" ) == 0 ){
        Nsigma = strtod( argv[++i],NULL );
      }
      else if ( strcmp( argv[i],"-pause" ) == 0 ){
        Pause = true;
      }
      else if ( strcmp( argv[i],"-cycle" ) == 0 ){
        Cycle = strtod( argv[++i],NULL );
      }
      else if ( strcmp( argv[i],"-freq" ) == 0 ){
        Freq = strtod( argv[++i],NULL );
      }
      else if ( strcmp( argv[i],"-int" ) == 0 ){
        Interval = strtol( argv[++i],NULL,0 );
      }
      else if ( strcmp( argv[i],"-usleep" ) == 0 ){
        Usleep = strtol( argv[++i],NULL,0 );
      }
      else if ( strcmp( argv[i],"-trig" ) == 0 ){
        TriggerBit = strtol( argv[++i],NULL,0 );
      }
      else if ( strcmp( argv[i],"-rot" ) == 0 ){
        RotAngle = strtod( argv[++i],NULL );
      }
      else if ( strcmp( argv[i],"-psfile" ) == 0 ){
        Psfile = argv[++i];
      }
      else if ( strcmp( argv[i],"-tracking" ) == 0 ){
        Tracking = true;
      }
      else if ( strcmp( argv[i],"-noscreen" ) == 0 ){
        WkType = 0;
      }
      else if ( strcmp( argv[i],"-d" ) == 0 ){
        Debug = 1;
      }
      else{
        std::cout << " unknown option : " << argv[i] << std::endl;
        usage( argc,argv );
        return( 1 );
      }
    }
    else{
      std::cout << " unknown option : " << argv[i] << std::endl;
      usage( argc,argv );
      return( 1 );
    }
  }
  // freq
  if ( Freq <= 0 )
    Freq = 1;
    
  // iflist
  for( std::list<std::string>::iterator i = iftmp.begin();
       i != iftmp.end(); i++ ){
    if ( strchr( i->c_str(),'%' ) ){
      for( int rn=runFrom; rn<=runTo; rn++ ){
        char  s[1024];
        sprintf( s,i->c_str(),rn );
        iflist.push_back( std::string( s ) );
      }
    }
    else{
      iflist.push_back( *i );
    }
  }
  
  // call user init routine
  user_init();
  
  // main loop
  if ( iflist.size() > 0 ){
    // read from file
    for( std::list<std::string>::iterator f = iflist.begin();
         f != iflist.end(); f++ ){
      // open data file
      MidasFileIO  mfile( *f,O_RDONLY|O_LARGEFILE );
      if ( mfile.good() ){
        std::cout << "Opening input file : " << *f << std::endl;
        
        // event bank
        EVENT_HEADER*  pheader = mfile.bank()->eventHeader();
        void*          pevent  = mfile.bank()->eventBuffer();
        
        // Begin_of_run record
        int  run_number = -1;
        if ( mfile.readBOR() ){
          run_number = pheader->serial_number;
          std::cout << "Run = " << run_number << std::endl;
          e391det.setRunNumber( run_number );
          // get DB for this run
          e391det.getDB();
          //e391det.getDB( 1 );
          
          // call user begin_of_run routine
          user_begin_of_run( run_number );
        }
        
        // Process Event
        mfile.skipEvent( 0,Skipevent );
        for( int i=0; i<Nevent || Nevent<0; i++ ){
          int  id = mfile.readEvent(0);
          if ( (id & 0x8000) == 0x8000 ){
            // Special event (BOR,EOR,MES)
            if ( id == EVENTID_BOR ){
              run_number = pheader->serial_number;
              std::cout << "Run=" << run_number << std::endl;
              e391det.setRunNumber( run_number );
              // get DB for this run
              e391det.getDB();
              //e391det.getDB( 1 );
              
              // call user begin_of_run routine
              user_begin_of_run( run_number );
            }
            else if ( pheader->event_id == EVENTID_EOR ){
              // call user end_of_run routine
              user_end_of_run( run_number );
            }
          }
          else if ( id > 0 ){
            e391det.setEvent( pheader, pevent );
            if ( e391det.exist( "EVNT" ) ){
              // built data
              // trigger bit
              int  trig = e391det.triggerBit();
              // unpack event
              if ( TriggerBit == 0 || (trig & TriggerBit) ){
                e391det.unpack();
                int  spill_number = e391det.spillNumber();
                int  event_number = e391det.eventNumber();
                // call user analysis routine
                user_process_event(run_number,spill_number,event_number,trig);
              }
            }
            else{
              // raw data
              int  nev = 0;
              while( 1 ){
                if ( e391det.unpack( ++nev ) == 0 ){
                  // trigger bit
                  int  trig = e391det.triggerBit(nev);
                  if ( TriggerBit == 0 || (trig & TriggerBit) ){
                    int  spill_number = e391det.spillNumber();
                    // call user analysis routine
                    user_process_event( run_number,spill_number,nev,trig );
                  }
                }
                else{
                  break;
                }
              }
            }
          }
          else{
            // end of file
            break;
          }
        }
        
      }
      else{
        std::cout << "No such file : " << *f << std::endl;
      }
      
    }
    
  }
  ////
  else{
    // connect to experiment
    int  status = cm_connect_experiment( exphost,"",
                                         "EventDisplay",NULL );
    if ( status != CM_SUCCESS )
      return( -1 );
    
    /* open ODB structures */
    RUNINFO_STR(runinfo_str);
    HNDLE hDB, hKey;
    char  str[80];
    cm_get_experiment_database(&hDB, NULL);
    db_create_record(hDB, 0, "/Runinfo", strcomb(runinfo_str));
    db_find_key(hDB, 0, "/Runinfo", &hKey);
    if ( db_open_record(hDB,hKey,&runinfo, sizeof(runinfo),MODE_READ,NULL,NULL)
         != DB_SUCCESS ){
      std::cout << "cannot open /Runinfo." << std::endl;
      return 0;
    }

    HNDLE  hbuf;
    int  request_id;
    bm_open_buffer( "SYSTEM",EVENT_BUFFER_SIZE,&hbuf );
    bm_request_event( hbuf, 1, TRIGGER_ALL, GET_SOME,
                      &request_id, user_process_event_exp );
    do{
      status = cm_yield( 1000 );
    } while( status != RPC_SHUTDOWN &&
             status != SS_ABORT );

    cm_disconnect_experiment();
  }

  // call user term routine
  user_term();

  //
  return( 0 );
}

