/* -*- C -*-

   hpss_Bufapi.c

   Author:  Mitsuhiro YAMAGA
   Created: Wed Dec 11 22:57:06 JST 2002

   $Id: hpss_Bufapi.c,v 1.1 2002/12/11 14:12:33 yamaga Exp $

   Revision History

   $Log: hpss_Bufapi.c,v $
   Revision 1.1  2002/12/11 14:12:33  yamaga
   Imported package.


*/
/*            HPSS_BUFAPI    M.AOKI Version

   Based on hpss_Bufapi.c provided by IBM.  M.Aoki modified hpss_bufseek(),
   hpss_bufflush() and others in order to improve performance.

*/

static char rcsid_hpss_Bufapi[] = "@(#) $Id: hpss_Bufapi.c,v 1.1 2002/12/11 14:12:33 yamaga Exp $";


#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include "hpss_types.h"
#include "hpss_api.h"
#include "u_signed64.h"

#ifndef DBG
#define DBG 0
#endif

#define MegaBytes  (1024 * 1024)
#define STAGE_BUFFER_SIZE   (8 * MegaBytes)

#define NUM_FILES   256

typedef enum {
  FUNC_NONE,
  FUNC_READ,
  FUNC_WRITE
} func_types;

typedef enum {
  BUFF_NONE,
  BUFF_CLEAN,
  BUFF_DIRTY
} buff_status;

typedef struct {
  long long global_position;
  char *stage_buffer;
  long long stage_size;
  long long stage_offset;
  long long local_offset;
  func_types cur_function;
  long long buffer_size;
  buff_status buffer_status;
  char* filename;
  int   flags;
  int   mode;
} static_data;

static static_data files[NUM_FILES];


/*---------------------------------------*/
/* Initialize the stage buffer variables */
/*---------------------------------------*/
int hpss_Bufcreate (char *filename,
		    int flags,
		    int mode,
		    unsigned long long filesize)
{
  hpss_cos_hints_t      hints_in;
  hpss_cos_hints_t      hints_out;
  hpss_cos_priorities_t hints_pri; 
  long long global_position = 0;
  long long stage_size = 0;
  long long stage_offset = 0;
  long long local_offset = 0;
  long long file_offset = 0;
  char *stage_buffer;
  char *fname;
  int fd;

  if (DBG) printf("s - hpss_Bufcreate, file %s, size = %llu\n", 
		  filename, filesize);

  memset(&hints_in, 0x00, sizeof(hints_in));
  memset(&hints_pri, 0x00, sizeof(hints_pri));
  hints_in.MinFileSize = cast64m(filesize);
  hints_in.MaxFileSize = cast64m(filesize);

  if ((stage_buffer = (char *)malloc(STAGE_BUFFER_SIZE)) == (char *)NULL) {
    /* error */
    perror( "hpss_Bufcreate : malloc failed for stage_buffer" );
    return -1;
  }	
  if ( ( fname = (char *)malloc( strlen(filename)+1 ) ) != NULL ){
    strcpy( fname,filename );
  }
  else{
    /* error */
    perror( "hpss_Bufcreate : malloc failed for filename" );
    return -1;
  }

  flags |= O_CREAT;
  /*   flags |= O_LARGEFILE;  */
  fd = hpss_Open(filename, flags, mode, &hints_in, &hints_pri, &hints_out);
  if (DBG) printf("e - hpss_Bufcreate, fd = %d\n", fd);
  if ( fd >= 0 ) {
    files[fd].global_position = 0;
    files[fd].stage_buffer = stage_buffer;
    files[fd].stage_size = 0;
    files[fd].stage_offset = 0;
    files[fd].local_offset = 0;
    files[fd].cur_function = FUNC_NONE;
    files[fd].buffer_size = STAGE_BUFFER_SIZE;
    files[fd].buffer_status = BUFF_NONE;
    files[fd].filename = fname;
    files[fd].flags = flags;
    files[fd].mode = mode;
  }
  else{
    free( stage_buffer );
    free( fname );
  }

  return fd;
}


/*---------------------------------------*/
/* Initialize the stage buffer variables */
/*---------------------------------------*/
int hpss_Bufopen (char *filename,
		  int flags,
		  int mode)
{
  hpss_cos_hints_t      hints_in;
  hpss_cos_hints_t      hints_out;
  hpss_cos_priorities_t hints_pri; 
  long long global_position = 0;
  long long stage_size = 0;
  long long stage_offset = 0;
  long long local_offset = 0;
  long long file_offset = 0;
  char *stage_buffer;
  char *fname;
  int fd;

  if (DBG) printf("s - hpss_Bufopen, file %s flags 0x%x mode %o\n",
                  filename,flags,mode);

  if ((stage_buffer = (char *)malloc(STAGE_BUFFER_SIZE)) == (char *)NULL) {
    /* error */
    perror( "hpss_Bufopen : malloc failed for stage_buffer" );
    return -1;
  }	
  if ( ( fname = (char *)malloc( strlen(filename)+1 ) ) != NULL ){
    strcpy( fname,filename );
  }
  else{
    /* error */
    perror( "hpss_Bufopen : malloc failed for filename" );
    return -1;
  }

  /*   flags |= O_LARGEFILE;  */
  fd = hpss_Open(filename, flags, mode, NULL, NULL, &hints_out);
  if ( fd >= 0 ) {
    files[fd].global_position = 0;
    files[fd].stage_buffer = stage_buffer;
    files[fd].stage_size = 0;
    files[fd].stage_offset = 0;
    files[fd].local_offset = 0;
    files[fd].cur_function = FUNC_NONE;
    files[fd].buffer_size = STAGE_BUFFER_SIZE;
    files[fd].buffer_status = BUFF_NONE;
    files[fd].filename = fname;
    files[fd].flags = flags;
    files[fd].mode = mode;
  }
  else{
    free( stage_buffer );
    free( fname );
  }

  if (DBG) printf("e - hpss_Bufopen, fd = %d\n", fd);
  return fd;
}


/*---------------------------------------*/
/* read from HPSS into the stage buffer  */
/*---------------------------------------*/
int hpss_Bufread (int fd,
		  void *buffer,
		  size_t size)
{
  char *buf_ptr;
  int rc;
  long long current_position;
  u_signed64 offset_in;
  u_signed64 offset_out;

  if (DBG) printf("s - hpss_Bufread, %d bytes to fd %d\n", size, fd);

  /* recalibrate local_offset */
  files[fd].local_offset = files[fd].global_position - files[fd].stage_offset;

  /* Check if it is outside of the current stage buffer. */
  if ( files[fd].stage_size > 0 ) {
    if ( files[fd].local_offset < 0 ||
	 files[fd].local_offset > files[fd].stage_size-1 ) {

      /* flush out any outstanding data */
      rc = hpss_Bufflush(fd);
      if ( rc == -1 ) return rc;

    } else if ( files[fd].local_offset+size-1 > files[fd].stage_size-1 ) { 
      /* make sure the buffer is large enough */
      /* flush out any outstanding data before modify the stage buffer */
      rc = hpss_Bufflush(fd);
      if ( rc == -1 ) return rc;
    }
  }

  /* if we don't have any data in the stage buffer, read some in */
  if (DBG) printf("    stage_size: %lld\n", files[fd].stage_size);
  if ( files[fd].stage_size == 0 ) {
    /* make sure the buffer is large enough */
    if ( files[fd].buffer_size < size ) {
      if ((buf_ptr = (char *)realloc(files[fd].stage_buffer, size)) == 
	  (char *)NULL) {
	/* error */
	return -1;
      }	
      files[fd].stage_buffer = buf_ptr;
      files[fd].buffer_size = size;
    }

    /* Set the file position */
    CONVERT_LONGLONG_TO_U64((long long) files[fd].global_position, offset_in);
    rc = hpss_SetFileOffset(fd, offset_in, SEEK_SET,
			    HPSS_SET_OFFSET_FORWARD,
			    &offset_out);
    if (rc != HPSS_E_NOERROR) return(-1);
    CONVERT_U64_TO_LONGLONG(offset_out, current_position);
    files[fd].stage_offset = current_position;
    files[fd].global_position = current_position;

    if (DBG) printf("    read %lld bytes from HPSS for fd: %d\n", 
		    files[fd].buffer_size, fd);
    files[fd].stage_size = hpss_Read(fd, files[fd].stage_buffer, 
				     files[fd].buffer_size);
    if ( files[fd].stage_size <= 0 ) {
      if (DBG) printf("e - hpss_Bufread returned %lld\n",files[fd].stage_size);
      return files[fd].stage_size;
    }
    if (files[fd].stage_size < size) size = files[fd].stage_size;
    if (DBG) printf("    size: %d\n", size);
    files[fd].local_offset = 0;
    files[fd].buffer_status = BUFF_CLEAN;
  }
 
  /* copy the stage buffer to the user buffer */
  if (size > (files[fd].stage_size - files[fd].local_offset))
    size = files[fd].stage_size - files[fd].local_offset;
  if (DBG) printf("    memcpy size: %d\n", size);
  memcpy(buffer, files[fd].stage_buffer + files[fd].local_offset, size); 

  /* Update the stage buffer status */
  files[fd].local_offset += size;
  files[fd].global_position += size;
  files[fd].cur_function = FUNC_READ;

  if (DBG) printf("e - hpss_Bufread, size = %d\n", size);
  return size;
}

/*---------------------------------------*/
/* write from the stage buffer into HPSS */
/*---------------------------------------*/
int
hpss_Bufwrite(int fd, void *buffer, size_t size)
{
  char *buf_ptr;
  int rc;

  if (DBG) printf("s - hpss_Bufwrite, %d bytes to fd %d\n", size, fd);

  /* recalibrate local_offset */
  files[fd].local_offset = files[fd].global_position - files[fd].stage_offset;

  /* Check if it is outside of the current stage buffer. */
  if ( files[fd].local_offset < 0 ||
       files[fd].local_offset > files[fd].buffer_size-1 ) {

    /* flush out any outstanding data */
    rc = hpss_Bufflush(fd);
    if ( rc == -1 ) return rc;
    files[fd].stage_offset = files[fd].global_position;
  }

  /* make sure the buffer is large enough */
  if ( files[fd].local_offset+size-1 > files[fd].buffer_size-1 ) { 
    /* flush out any outstanding data before modify the stage buffer */
    rc = hpss_Bufflush(fd);
    if ( rc == -1 ) return rc;
    files[fd].stage_offset = files[fd].global_position;

    if ( size > files[fd].buffer_size ) {
      if ((buf_ptr = (char *)realloc(files[fd].stage_buffer, size)) == 
	  (char *)NULL) {
	/* error */
	return -1;
      }	
      files[fd].stage_buffer = buf_ptr;
      files[fd].buffer_size = size;
    }	
  }

  /* copy the buffer to the stage buffer */
  memcpy(files[fd].stage_buffer + files[fd].local_offset, buffer, size); 
  files[fd].buffer_status = BUFF_DIRTY;
  files[fd].local_offset += size;
  files[fd].global_position += size;
  if ( files[fd].local_offset > files[fd].stage_size ) 
    files[fd].stage_size = files[fd].local_offset;
  files[fd].cur_function = FUNC_WRITE;

  if (DBG) printf("e - hpss_Bufwrite, size = %d\n", size);
  return size;
}

/*----------------*/
/* close the file */
/*----------------*/
int
hpss_Bufclose(int fd)
{
  int rc;

  if (DBG) printf("s - hpss_Bufclose, fd %d\n", fd);

  /* flush out any outstanding data */
  rc = hpss_Bufflush(fd);
  if ( rc == -1 ) return rc;

  rc = hpss_Close(fd);

  if ( rc >= 0 ){
    if ( files[fd].stage_buffer ) {
      free(files[fd].stage_buffer);
      files[fd].stage_buffer = NULL;
    }
    if ( files[fd].filename ) {
      free(files[fd].filename);
      files[fd].filename = NULL;
    }

    files[fd].cur_function = FUNC_NONE;
  }

  if (DBG) printf("e - hpss_Bufclose : rc=%d\n",rc);
  return ( rc );
}

/*-----------------*/
/* delete the file */
/*-----------------*/
int
hpss_Bufdelete(char *filename)
{
  int rc;
  if (DBG) printf("s - hpss_Bufdelete, filename %s\n", filename);
  rc = hpss_Unlink(filename);
  if (DBG) printf("e - hpss_Bufdelete\n");
  return (rc);
}

/*---------------------------------*/
/* seek to a possition in the file */
/*---------------------------------*/
off_t
hpss_Bufseek(int fd, off_t len, int pos)
{
  unsigned long off1, off2;
  off_t offset;
  int direction;
  int rc;
  struct stat stat;

  /* This routine will only update the files[fd].golbal_position
     Actuall action should be taken by hpss_Bufread or hpss_Bufwrite and
     hpss_Bufflush. */


  if ( pos == SEEK_SET ) {
    files[fd].global_position = len;
  } else if ( pos == SEEK_CUR ) {
    files[fd].global_position += len;
  } else if ( pos == SEEK_END ) {
    /* Get the file size */
    rc = hpss_Fstat(fd, &stat);
    if ( rc < 0 ) {
      return(-1);  /* errno is supposed to be set by hpss_Fstat */
    }
    files[fd].global_position = stat.st_size + len;
  }

  return((off_t)files[fd].global_position);
}


/*---------------------------------*/
/* flush the stage buffer          */
/*---------------------------------*/
int
hpss_Bufflush(int fd)
{
  u_signed64 offset_in;
  u_signed64 offset_out;
  long long eof_position;
  struct stat stat;
  char *tmp_buffer;
  int rc = 0;
  ssize_t written_size;

  if (DBG) printf("s - hpss_Bufflush, local_offset = %d\n", 
		  files[fd].local_offset);

  /* flush only if the dirty flag is on */
  if ( files[fd].buffer_status == BUFF_DIRTY ) {
    if ( files[fd].stage_size > 0 ) {
      /* Set the file position */
      /* NOTE: hpss_SetFileOffset does not allow to move the pointer
	 beyond the file size.  Especially if the file is a new file,
	 it return with HPSS_EAGAIN error. */
      rc = hpss_Fstat(fd, &stat);
      if ( rc < 0 ) return(-1);

      /* if the file is empty, then it should not use hpss_SetFileOffset */
      if ( stat.st_size == 0 ) {
	if ( files[fd].stage_offset > 0 ) {
	  /* stage_offset > 0, then pad bytes  */
	  if ( (tmp_buffer = (char *)malloc(files[fd].stage_offset))
	       == (char *)NULL) {
	    /* error */
	    return -1;
	  }	
	  written_size = hpss_Write(fd, tmp_buffer, files[fd].stage_offset);
	  if (written_size != files[fd].stage_offset) return(-1);
	  free(tmp_buffer);
	}
      } else {
	if ( files[fd].stage_offset <= stat.st_size-1 ) {
	  /* If the stage_offset is within the existing file size */
	  CONVERT_LONGLONG_TO_U64((long long) files[fd].stage_offset, offset_in);
	  rc = hpss_SetFileOffset(fd, offset_in, SEEK_SET,
				  HPSS_SET_OFFSET_FORWARD,
				  &offset_out);
	  if (rc != HPSS_E_NOERROR) return(-1);
	} else {
	  /* If the stage_offset is beyond the existing file size, pad bytes */
	  CONVERT_LONGLONG_TO_U64((long long) 0, offset_in);
	  rc = hpss_SetFileOffset(fd, offset_in, SEEK_END,
				  HPSS_SET_OFFSET_FORWARD,
				  &offset_out);
	  if (rc != HPSS_E_NOERROR) return(-1);
	  CONVERT_U64_TO_LONGLONG(offset_out, eof_position);

	  if ( files[fd].stage_offset > eof_position ) {
	    if ( (tmp_buffer = (char *)malloc(files[fd].stage_offset-eof_position))
		 == (char *)NULL) {
	      /* error */
	      return -1;
	    }	
	    written_size = hpss_Write(fd, tmp_buffer, files[fd].stage_offset-eof_position);
	    if (written_size != files[fd].stage_offset-eof_position) return(-1);
	    free(tmp_buffer);
	  }
	}
      }

      written_size = hpss_Write(fd, files[fd].stage_buffer, files[fd].stage_size);
      if (DBG) printf("    flushed buffer, stage_size = %lld\n", 
		      files[fd].stage_size);
      if (written_size != files[fd].stage_size) return(-1);

    }
  }

  files[fd].buffer_status = BUFF_NONE;
  files[fd].stage_size = 0;
  files[fd].stage_offset = 0;
  files[fd].local_offset = 0;
  files[fd].cur_function = FUNC_NONE;

  if (DBG) printf("e - hpss_Bufflush, %d\n", rc);
  return rc;
}



/*---------------------------------*/
/* Change the stage buffer size    */
/*---------------------------------*/
int
hpss_Bufsize(int fd, int size)
{
  char *buf_ptr;
  int rc;

  if ( size == 0 ) {
    return files[fd].buffer_size;
  } else {
    /* flush out any outstanding data before modify the stage buffer */
    rc = hpss_Bufflush(fd);
    if ( rc == -1 ) return rc;
    
    if ((buf_ptr = (char *)realloc(files[fd].stage_buffer, size)) == 
	(char *)NULL) {
      /* error */
      return -1;
    }	
    files[fd].stage_buffer = buf_ptr;
    files[fd].buffer_size = size;
  }
}


/*---------------------------------*/
/* reopen file                     */
/*---------------------------------*/
int
hpss_BufreopenAll( int nhfd, const int hfdlist[], int newhfdlist[] )
{
  static_data  bkfiles[NUM_FILES];
  int  flags;
  int  oldhfd,newhfd,idx,rc,i,j;

  if (DBG) printf( "s - hpss_BufreopenAll\n" );
  
  for( i=0;i<nhfd;i++ ){
    oldhfd = hfdlist[i];
    bkfiles[oldhfd].global_position = files[oldhfd].global_position;
    bkfiles[oldhfd].stage_buffer    = files[oldhfd].stage_buffer;
    bkfiles[oldhfd].stage_size      = files[oldhfd].stage_size;
    bkfiles[oldhfd].stage_offset    = files[oldhfd].stage_offset;
    bkfiles[oldhfd].local_offset    = files[oldhfd].local_offset;
    bkfiles[oldhfd].cur_function    = files[oldhfd].cur_function;
    bkfiles[oldhfd].buffer_size     = files[oldhfd].buffer_size;
    bkfiles[oldhfd].buffer_status   = files[oldhfd].buffer_status;
    bkfiles[oldhfd].filename        = files[oldhfd].filename;
    bkfiles[oldhfd].flags           = files[oldhfd].flags;
    bkfiles[oldhfd].mode            = files[oldhfd].mode;
    if (DBG) printf( "    idx=%d oldhfd=%d\n",i,oldhfd );
  }
  
  for( i=0;i<nhfd;i++ ){
    oldhfd = hfdlist[i];
    rc = hpss_Close( oldhfd );
    if (DBG) printf( "    closed oldhfd=%d rc=%d err=%d\n",
                     oldhfd,rc,errno );
  }
  
  /* reset */
  hpss_ClientAPIReset();
  
  /* reopen */
  for( i=0;i<nhfd;i++ ){
    if ( (flags = bkfiles[i].flags) & (O_WRONLY|O_RDWR) )
      flags = (flags | O_APPEND) & ~(O_EXCL | O_TRUNC);
    newhfd = hpss_Bufopen( bkfiles[i].filename, flags, bkfiles[i].mode );
    if ( newhfd >= 0 ){
      newhfdlist[i] = newhfd;
      if (DBG) printf( "    idx=%d oldhfd=%d newhfd=%d\n",
                       i,hfdlist[i],newhfdlist[i] );
    }
    else{ /* failed */
      rc = newhfd; /* return code of hpss_Bufopen */
      if (DBG) printf( "    failed to reopen file %s rc=%d err=%d\n",
                       bkfiles[i].filename,rc,errno );
      for( j=i-1;j>=0;j-- ){
        newhfd = newhfdlist[j];
        free( files[newhfd].stage_buffer );
        free( files[newhfd].filename );
      }
      /* restore */
      for( j=0;j<nhfd;j++ ){
        oldhfd = hfdlist[j];
        files[oldhfd].global_position = bkfiles[oldhfd].global_position;
        files[oldhfd].stage_buffer    = bkfiles[oldhfd].stage_buffer;
        files[oldhfd].stage_size      = bkfiles[oldhfd].stage_size;
        files[oldhfd].stage_offset    = bkfiles[oldhfd].stage_offset;
        files[oldhfd].local_offset    = bkfiles[oldhfd].local_offset;
        files[oldhfd].cur_function    = bkfiles[oldhfd].cur_function;
        files[oldhfd].buffer_size     = bkfiles[oldhfd].buffer_size;
        files[oldhfd].buffer_status   = bkfiles[oldhfd].buffer_status;
        files[oldhfd].filename        = bkfiles[oldhfd].filename;
        files[oldhfd].flags           = bkfiles[oldhfd].flags;
        files[oldhfd].mode            = bkfiles[oldhfd].mode;
      }
      if (DBG) printf( "e - hpss_BufreopenAll : nfiles=%d return=%d\n",
                       nhfd,rc );
      return( rc );
    }
  }
  
  /* restore information */
  for( i=0;i<nhfd;i++ ){
    oldhfd = hfdlist[i];
    newhfd = newhfdlist[i];
    free( files[newhfd].stage_buffer );
    free( files[newhfd].filename );
    files[newhfd].global_position = bkfiles[oldhfd].global_position;
    files[newhfd].stage_buffer    = bkfiles[oldhfd].stage_buffer;
    files[newhfd].stage_size      = bkfiles[oldhfd].stage_size;
    files[newhfd].stage_offset    = bkfiles[oldhfd].stage_offset;
    files[newhfd].local_offset    = bkfiles[oldhfd].local_offset;
    files[newhfd].cur_function    = bkfiles[oldhfd].cur_function;
    files[newhfd].buffer_size     = bkfiles[oldhfd].buffer_size;
    files[newhfd].buffer_status   = bkfiles[oldhfd].buffer_status;
    files[newhfd].filename        = bkfiles[oldhfd].filename;
    files[newhfd].flags           = bkfiles[oldhfd].flags;
    files[newhfd].mode            = bkfiles[oldhfd].mode;
    if (DBG) printf( "    copy idx=%d oldhfd=%d newhfd=%d buf=%lld\n",
                     i,oldhfd,newhfd,files[newhfd].stage_size );
  }

  if (DBG) printf( "e - hpss_BufreopenAll : nfiles=%d\n",nhfd );
  
  return( nhfd );
}


